/*
 * Copyright (c) 2009-2010 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef COLORFILTERING_IPOL_H
#define COLORFILTERING_IPOL_H


/**
* \brief Data structure containing the parameters of the filtering algorithm 
*
* \param rneigh      distance defining the neighborhood of a color point 
*                    (neighbors are RGB points at L1-distance below rneigh)
* \param rsim        distance defining a small neighborhood around each color 
*                    point, it is used to accelerate the algorithm. It must be 
*                    much smaller than rneigh, typically rsim=rneigh/3 
* \param nneighmin   minimum number of neighbors needed to compute a local 
*                    line or plane in the vicinity of a color point 
* \param eitmax      stop criterion of the algorithm. The algorithm is 
*                    iterated until the average difference of corresponding
*                    colors in two consecutive iterations is below eitmax  
* \param option      project points to lines (option=1) or planes (option=2)  
* \param ignore000   if set to 1, do not take into account pixels with (0,0,0) 
*                    RGB value (eventually discarded for computation in a 
*                    preprocessing step)  
*
*/
struct paramColorFilter {
  int rneigh; //distance defining the neighborhood of a color point 
              //(neighbors are RGB points at L1-distance below rneigh)
  int rsim; //distance defining a small neighborhood around each color point, 
            //it is used to accelerate the algorithm. 
            //It must be much smaller than rneigh, typically rsim=rneigh/3 
  int nneighmin; //minimum number of neighbors needed to compute a local line 
                 //or plane in the vicinity of a color point 
  float eitmax; //stop criterion of the algorithm. The algorithm is iterated 
                //until the average difference of corresponding colors in two 
                //consecutive iterations is below eitmax 
  unsigned char option; //project points to lines (option=1) 
                        //or planes (option=2) 
  unsigned char ignore000; //if set to 1, do not take into account pixels with 
                           //(0,0,0) RGB value (eventually discarded for
                           //computation in a preprocessing step) 
};

struct paramColorFilter *new_colorfiltering_parameters(int rneigh, int rsim, 
                                                       int nneighmin, 
                                                       float eitmax,
                                                       unsigned char option, 
                                                      unsigned char ignore000);

void delete_colorfiltering_parameters(struct paramColorFilter *param);

//option=1 -> project to line 
//option=2 -> project to plane 
//option=3 -> project to line if weight1 >= weightMin, else project to plane
int colorfiltering(unsigned char *R, unsigned char *G, unsigned char *B, 
                   unsigned char *outR, unsigned char *outG, 
                   unsigned char *outB,
                   int w, int h, struct paramColorFilter *param);

float ***RGBdensities(unsigned char *R, unsigned char *G, unsigned char *B, 
                      int w, int h, int rdst, float hdst, float &maxdst);

void delete_RGBdensities(float ***dsts);

void removeisolatedRGB(unsigned char *R, unsigned char *G, unsigned char *B, 
                       unsigned char *outR, unsigned char *outG, 
                       unsigned char *outB,
                       int w, int h, int r, int nmin, int &nRGB, 
                       int &nisolatedRGB, 
                       int &nisolatedpixels);

#endif

